/* トラ技2023年3月号ダウンロードサービス
 * 「ESP32 のアナログ性能徹底調査」 
 *
 * Target: M5Stack ATOM Lite(https://docs.m5stack.com/en/core/atom_lite)
 *     ADC入力ピン G33
 *     DAC出力ピン G25
 * 
 * 1. PC との間に USBケーブルを接続する。
 * 2. PC からは、USBシリアル通信を行う。ボーレート 9600bps
 * 3. PC からのコマンドは、ASCII 通信とし、ターミナル文字は CR (0x0d)
 * 
 * コマンド:
 * 0    ADCの減衰率を 0dB に設定する。M5Stackからの応答は無い。
 * 1    ADCの減衰率を 2.5dB に設定する。M5Stackからの応答は無い。
 * 2    ADCの減衰率を 6dB に設定する。M5Stackからの応答は無い。
 * 3    ADCの減衰率を 11dB に設定する。M5Stackからの応答は無い。
 * a    ADCのバイナリ値を問合わせる。M5Stackからの応答は、10進数ASCII文字で、ターミナル文字 CR
 * v    ADCの生値から、内部校正パラメータを適用して得られた電圧値[mV]を問い合わせる。
 *      M5Stackからの応答は、10進数ASCII文字で、ターミナル文字 CR
 * dxx  DACのバイナリ値を設定する。バイナリ値は16進数2桁(00-FF) で表記する。
 * 
 * 例(ターミナル文字 CR は省略)
 * ----------
 * PCからの通信     M5Stackからの応答    意味
 * 0              (なし)              ADCの減衰率を 0dB に設定する。
 * d3D            (なし)              DACのバイナリ値を 61 (=0x3d) に設定する。
 * a              1232                ADCのバイナリ値を問合せ、1232 と応答された。(最小0 - 最大4095)
 */
#define ADC_PIN (33)
#define DAC_PIN (25)
#define STACK_SIZE (256)

uint8_t  command_stack[STACK_SIZE];
uint16_t command_index = 0;

int16_t hex2bin(uint8_t upper, uint8_t lower){
  /* 2桁の16進数文字を、数値に変換
   * 
   * Parameters
   * -----
   * upper
   *    2桁の16進数文字表現の上桁 ('0'-'F')
   * lower:
   *    2桁の16進数文字表現の下桁 ('0'-'F')
   * 
   * Return
   * -----
   * 数値に変換された値 ( upper='1'/lower='C' => 28 )
   * 異常入力時は -1
   */
  int16_t val;
  
  if(upper >= '0' && upper <= '9'){
    val = ((int16_t)upper-'0') << 4;
  }
  else if(upper >= 'a' && upper <= 'f'){
    val = ((int16_t)upper-'a'+10) << 4;
  }
  else if(upper >= 'A' && upper <= 'F'){
    val = ((int16_t)upper-'A'+10) << 4;
  }
  else{
    return(-1);
  }
  if(lower >= '0' && lower <= '9'){
    val |= (lower-'0');
  }
  else if(lower >= 'a' && lower <= 'f'){
    val |= (lower-'a'+10);
  }
  else if(lower >= 'A' && lower <= 'F'){
    val |= (lower-'A'+10);
  }
  else{
    return(-1);
  }
  return(val);
}

void do_command(const uint8_t* stack, uint16_t index){
  /* PC から送られたコマンドをデコードして、実行
   * 
   * Parameters
   * -----
   * stack
   *    コマンド文字列の先頭ポインタ (文字列ターミナル文字 \0 は不要)
   * index
   *    コマンド文字列の有効サイズ (ターミナル文字は含まない、実質コマンド文字数)
   */
  uint16_t adc_val;
  uint32_t adc_mv;
  int16_t  dac_val;
  
  if(index >= 1){
    switch(stack[0]){
      case '0':
        analogSetPinAttenuation(ADC_PIN, ADC_0db);
        break;
      case '1':
        analogSetPinAttenuation(ADC_PIN, ADC_2_5db);
        break;
      case '2':
        analogSetPinAttenuation(ADC_PIN, ADC_6db);
        break;
      case '3':
        analogSetPinAttenuation(ADC_PIN, ADC_11db);
        break;
      case 'a':
        adc_val = analogRead(ADC_PIN);
        Serial.print(adc_val);
        Serial.print('\r');
        break;
      case 'v':
        adc_mv = analogReadMilliVolts(ADC_PIN);
        Serial.print(adc_mv);
        Serial.print('\r');
        break;
      case 'd':
        dac_val = -1;
        if(index >= 3){
          dac_val = hex2bin(stack[1], stack[2]);
        }
        if(dac_val >= 0){
          dacWrite(DAC_PIN, (uint8_t)dac_val);
        }
        break;
      default:
        break;
    }
  }
}

void setup() {
  Serial.begin(9600);
}

void loop() {
  int16_t ser_in;
  
  if(Serial.available()){
    ser_in = Serial.read();
    if(ser_in == '\r'){ // CR
      do_command(command_stack, command_index);
      command_index = 0;
    }
    else if(ser_in == '\n'){  // LF
      // Do nothing
    }
    else{
      if(command_index < STACK_SIZE){
        command_stack[command_index++] = (uint8_t) ser_in;
      }
    }
  }
}
